module Merging where

import MLPPE
import Processes
import LPPE 
import Debug.Trace
import Auxiliary
import Simplify
import Expressions
import DataSpec
import Data.Set

mergeTauSummands :: MSpecification -> MSpecification
mergeTauSummands (MLPPE name params summands, initialState, dataSpec) = (MLPPE name params newSummands, initialState, dataSpec)
  where
    newSummands = mergeTauSummands2 params dataSpec summands summands

mergeTauSummands2 :: ProcessPars -> DataSpec -> [GeneralSummand] -> [GeneralSummand] -> [GeneralSummand]
mergeTauSummands2 pars dataspec []             _        = []
mergeTauSummands2 pars dataspec (summand:rest) summands | reduction = (changeMProbChoices (changeMNextState summand newNextState) newProbChoices):(mergeTauSummands2 pars dataspec rest summands)
                                                        | otherwise = summand:(mergeTauSummands2 pars dataspec rest summands)
  where
    condition      = makeConditionMLPPE summand pars
    enabled        = interactiveSummandsEnabledByCondition condition summands
    disabled       = interactiveSummandsDisabledByCondition dataspec condition summands
    nextSummand    = enabled!!0
    reduction      =    length enabled == 1
                     && length disabled == length [s | s <- summands, isInteractive s] - 1 
                     && isInteractive nextSummand 
                     && getMAction nextSummand == "tau"
                     && getMLocalPars nextSummand == []
    newProbChoices = removeDoubles (getMProbChoices summand ++ getMProbChoices nextSummand)
    newNextState   = Prelude.map (substituteInExpression [(fst (pars!!i), (getMNextState summand)!!i) | i <- [0..length pars - 1], not (elem (fst (pars!!i)) (Prelude.map fst3 newProbChoices))]) (getMNextState nextSummand)

interactiveSummandsEnabledByCondition :: Condition -> [GeneralSummand] -> [GeneralSummand]
interactiveSummandsEnabledByCondition condition summands = [s | s <- summands, isInteractive s, conditionImplies condition (getMCondition s)]

interactiveSummandsDisabledByCondition :: DataSpec -> Condition -> [GeneralSummand] -> [GeneralSummand]
interactiveSummandsDisabledByCondition dataspec condition summands = [s | s <- summands, isInteractive s, conditionImplies condition (simplifyExpression dataspec (Function "not" [getMCondition s]))]


removeDoubles []                       = []
removeDoubles ((name, typ, expr):rest) | double    = if typ == TypeRange 1 1 then removeDoubles rest else error ("Error: variable " ++ name ++ " defined twice")
                                       | otherwise = (name,typ,expr):(removeDoubles rest)
  where
    double = length [n | (n,t,e) <- rest, n == name] > 0